#ifndef MSP_CORE_THREAD_H_
#define MSP_CORE_THREAD_H_

#include <string>
#include "mspcore_api.h"
#include "noncopyable.h"

namespace Msp {

/**
Base class for threads.  To create a thread for some task, derive it from this
class and implement the main() function.  Note that threads are not
automatically started upon creation - you must manually call launch() instead.
This is to allow initializing variables of the derived class before the thread
is started.
*/
class MSPCORE_API Thread: private NonCopyable
{
private:
	struct Private;

	enum State
	{
		PENDING,
		RUNNING,
		FINISHED,
		KILLED,
		JOINED
	};

	Private *_priv = nullptr;
	std::string _name;
	State _state = PENDING;

protected:
	Thread(const std::string & = std::string());
public:
	virtual ~Thread();

	const std::string &get_name() const { return _name; }

	/** Indicates whether the thread has finished running. */
	bool is_finished() { return _state>=FINISHED; }

	/** Waits for the thread to exit.  Calling this from the thread will cause a
	deadlock. */
	void join();

	/** Violently terminates the thread.  This should only be used as a last
	resort, as the thread gets no chance to clean up. */
	void kill();

protected:
	/** Starts the thread.  Can only be called once for each Thread instance. */
	void launch();

private:
	void platform_join();
	void platform_kill();
	void platform_launch();
	void platform_setname();

protected:
	virtual void main() = 0;
};

} // namespace Msp

#endif
