#ifndef MSP_STRINGS_LEXICALCAST_H_
#define MSP_STRINGS_LEXICALCAST_H_

#include <sstream>
#include <string>
#include <stdexcept>
#include <msp/core/meta.h>
#include <msp/core/mspcore_api.h>
#include "fmt.h"

namespace Msp {

/**
Thrown for errors in lexical conversions.
*/
class MSPCORE_API lexical_error: public std::runtime_error
{
public:
	lexical_error(const std::string &w): runtime_error(w) { }
};


/**
Thrown when the format is unsuitable for the type being converted.
*/
class MSPCORE_API format_mismatch: public lexical_error
{
public:
	format_mismatch(const std::string &w): lexical_error(w) { }
};


/**
Helper class for lexical_cast to facilitate operator overloading.
*/
class MSPCORE_API LexicalConverter
{
private:
	Fmt fmt;
	bool filled = false;
	std::string buf;

public:
	LexicalConverter(const Fmt &f): fmt(f) { }
	LexicalConverter(const std::string &s, const Fmt &f): fmt(f), filled(true), buf(s) { }

	const Fmt &get_fmt() const { return fmt; }
	const std::string &get() const;
	void result(const std::string &);
};


MSPCORE_API void operator<<(LexicalConverter &, char);
MSPCORE_API void operator<<(LexicalConverter &, signed char);
MSPCORE_API void operator<<(LexicalConverter &, short);
MSPCORE_API void operator<<(LexicalConverter &, int);
MSPCORE_API void operator<<(LexicalConverter &, long);
MSPCORE_API void operator<<(LexicalConverter &, unsigned char);
MSPCORE_API void operator<<(LexicalConverter &, unsigned short);
MSPCORE_API void operator<<(LexicalConverter &, unsigned);
MSPCORE_API void operator<<(LexicalConverter &, unsigned long);
MSPCORE_API void operator<<(LexicalConverter &, long long);
MSPCORE_API void operator<<(LexicalConverter &, unsigned long long);
MSPCORE_API void operator<<(LexicalConverter &, bool);
MSPCORE_API void operator<<(LexicalConverter &, float);
MSPCORE_API void operator<<(LexicalConverter &, double);
MSPCORE_API void operator<<(LexicalConverter &, long double);
MSPCORE_API void operator<<(LexicalConverter &, const std::string &);
MSPCORE_API void operator<<(LexicalConverter &, const char *);
MSPCORE_API void operator<<(LexicalConverter &, const void *);

MSPCORE_API void operator>>(const LexicalConverter &, char &);
MSPCORE_API void operator>>(const LexicalConverter &, signed char &);
MSPCORE_API void operator>>(const LexicalConverter &, short &);
MSPCORE_API void operator>>(const LexicalConverter &, int &);
MSPCORE_API void operator>>(const LexicalConverter &, long &);
MSPCORE_API void operator>>(const LexicalConverter &, unsigned char &);
MSPCORE_API void operator>>(const LexicalConverter &, unsigned short &);
MSPCORE_API void operator>>(const LexicalConverter &, unsigned int &);
MSPCORE_API void operator>>(const LexicalConverter &, unsigned long &);
MSPCORE_API void operator>>(const LexicalConverter &, long long &);
MSPCORE_API void operator>>(const LexicalConverter &, unsigned long long &);
MSPCORE_API void operator>>(const LexicalConverter &, bool &);
MSPCORE_API void operator>>(const LexicalConverter &, float &);
MSPCORE_API void operator>>(const LexicalConverter &, double &);
MSPCORE_API void operator>>(const LexicalConverter &, long double &);
MSPCORE_API void operator>>(const LexicalConverter &, std::string &);

// Generic operators using stringstream

struct CheckFormattedOutput: Sfinae
{
	static std::ostream &s;
	template<typename T>
	static Yes f(int (*)[sizeof(s<<reinterpret_cast<const T &>(s))]);
	using Sfinae::f;
};

struct CheckFormattedInput: Sfinae
{
	static std::istream &s;
	template<typename T>
	static Yes f(int (*)[sizeof(s>>reinterpret_cast<T &>(s))]);
	using Sfinae::f;
};

template<typename T> struct HasFormattedOutput: Sfinae::Evaluate<CheckFormattedOutput, T> { };
template<typename T> struct HasFormattedInput: Sfinae::Evaluate<CheckFormattedInput, T> { };


template<typename T>
typename std::enable_if<HasFormattedOutput<T>::value>::type
operator<<(LexicalConverter &c, const T &v)
{
	std::ostringstream ss;
	ss << c.get_fmt() << v;
	c.result(ss.str());
}

template<typename T>
typename std::enable_if<HasFormattedInput<T>::value>::type
operator>>(const LexicalConverter &c, T &v)
{
	std::istringstream ss(c.get());
	ss.setf(std::ios_base::fmtflags(0), std::ios_base::skipws);
	ss >> v;
	if(ss.fail() || !ss.eof())
		throw lexical_error("conversion failure");
}

/**
Helper struct to provide partial template specialization.
*/
template<typename T, typename F>
struct LexicalCast;

template<typename T>
struct LexicalCast<T, std::string>
{
	static T cast(const std::string &s, const Fmt &f = Fmt())
	{
		LexicalConverter conv(s, f);
		T result;
		conv >> result;
		return result;
	}
};

template<typename F>
struct LexicalCast<std::string, F>
{
	static std::string cast(const F &v, const Fmt &f = Fmt())
	{
		LexicalConverter conv(f);
		conv << v;
		return conv.get();
	}
};

template<>
struct LexicalCast<std::string, std::string>
{
	static std::string cast(const std::string &v, const Fmt &f = Fmt())
	{
		LexicalConverter conv(f);
		conv << v;
		return conv.get();
	}
};

/** Perform a lexical conversion between a string and another type.  The source
type can normally be deduced by the compiler, so this can be used just like the
standard C++ casts.  A format may additionally be specified to force a specific
interpretation. */
template<typename T, typename F>
inline T lexical_cast(const F &v, const Fmt &f = Fmt())
{
	return LexicalCast<T, F>::cast(v, f);
}

} // namespace Msp

#endif
